import os
from os import path
import sys

from typing import List, Dict, Iterator, Sequence, Union
from .structures import Structures


def write_match_output(structures: Structures, odir_match: str) -> None:
    """
    Save Tregex's match output

    :param structures: an instance of Structures
    :param dir_match_output: where to save the match output
    """
    bn_input = path.basename(structures.ifile)
    bn_input_noext = path.splitext(bn_input)[0]
    subdir_match_output = path.join(odir_match, bn_input_noext).strip()
    if not path.isdir(subdir_match_output):
        # if not (exists and is a directory)
        os.makedirs(subdir_match_output)
    for structure in structures.to_query:
        if structure.matches:
            bn_match_output = bn_input_noext + "-" + structure.name.replace("/", "p") + ".matches"
            fn_match_output = path.join(subdir_match_output, bn_match_output)
            with open(fn_match_output, "w", encoding="utf-8") as f:
                f.write(structure.matches)


def write_freq_output(
    multi_structures: Union[Iterator[Structures], Sequence[Structures]],
    ofile_freq,
    oformat_freq: str,
) -> None:
    if oformat_freq == "csv":
        freq_output = Structures.fields
        for structures in multi_structures:
            freq_dict = structures.get_freqs()
            freq_output += "\n" + ",".join(str(freq) for freq in freq_dict.values())
    elif oformat_freq == "json":
        import json

        combined_freq_dict: Dict[str, List[Dict]] = {"Files": []}
        for structures in multi_structures:
            freq_dict = structures.get_freqs()
            combined_freq_dict["Files"].append(freq_dict)
        freq_output = json.dumps(combined_freq_dict)
    else:
        print(f"Unexpected output format: {oformat_freq}")
        sys.exit(1)

    if ofile_freq is not sys.stdout:
        with open(ofile_freq, "w", encoding="utf-8") as f:
            f.write(freq_output)
    else:
        ofile_freq.write(freq_output)



from collections import namedtuple
import email.message
import glob
import lzma
import os
import re
import shutil
import subprocess
from sys import maxsize, platform
import tarfile
import tempfile
from typing import Optional, Union
from urllib.error import URLError
import urllib.parse
import urllib.request
import zipfile

from .util import same_line_print
from .util import color_print
from .util import get_yes_or_no
from .util import SCAProcedureResult

_IS_WINDOWS = os.name == "nt"
_IS_DARWIN = platform == "darwin"
_UNPACK200 = "unpack200.exe" if _IS_WINDOWS else "unpack200"
_UNPACK200_ARGS = '-r -v -l ""' if _IS_WINDOWS else ""
if _IS_WINDOWS and os.environ.get("ProgramFiles") is not None:
    _TARGET_DIR = os.environ.get("ProgramFiles")
else:
    _USER_DIR = os.path.expanduser("~")
    _TARGET_DIR = os.path.join(_USER_DIR, ".local", "share")

OS = "windows" if _IS_WINDOWS else "mac" if _IS_DARWIN else platform
ARCH = "x64" if maxsize > 2**32 else "x32"

_TAR = ".tar"
_TAR_GZ = ".tar.gz"
_ZIP = ".zip"
_SEVEN_ZIP = ".7z"

_Path = namedtuple("_Path", "dir base name ext")
JAVA = "Java"
_JAVA_VERSION = "8"
STANFORD_PARSER = "Stanford Parser"
STANFORD_TREGEX = "Stanford Tregex"


class Implementation:
    OPENJ9 = "openj9"
    HOTSPOT = "hotspot"


class depends_installer:
    def __init__(self) -> None:
        self._URL_JAVA_TEMPLATE = (
            "https://api.adoptopenjdk.net/v3/binary/latest/{}/ga/{}/{}/jdk/{}/normal/adoptopenjdk"
        )
        self._URL_JAVA_TEMPLATE_CHINA = "https://mirrors.tuna.tsinghua.edu.cn/Adoptium/{}/jdk/{}/{}/"
        self._URL_STANFORD_PARSER = "https://downloads.cs.stanford.edu/nlp/software/stanford-parser-4.2.0.zip"
        self._URL_STANFORD_TREGEX = "https://downloads.cs.stanford.edu/nlp/software/stanford-tregex-4.2.0.zip"
        self.headers = {"User-Agent": "Mozilla/5.0"}

    def normalize_version(self, version: str) -> SCAProcedureResult:
        if re.search(r"^\d+$", version):
            return True, version
        match = re.search(r"^\d+\.(\d+)$", version)
        if match:
            return True, match.group(1)
        return False, f"Error: Unexpected version format: {version}."

    def get_java_download_url(
        self,
        version: str,
        operating_system: Optional[str] = None,
        arch: Optional[str] = None,
        impl: str = Implementation.HOTSPOT,
    ) -> SCAProcedureResult:
        sucess, err_msg = self.normalize_version(version)
        if not sucess:
            return sucess, err_msg
        else:
            version = err_msg  # type:ignore
        self.use_chinese_jdk_mirror = get_yes_or_no(
            "Do you want to download Java from a Chinese mirror site?"
            " If you are inside of China, you may want to use this for a faster network connection."
        )
        if self.use_chinese_jdk_mirror in ("n", "N"):
            return True, self._URL_JAVA_TEMPLATE.format(version, operating_system, arch, impl)
        else:
            index_url = self._URL_JAVA_TEMPLATE_CHINA.format(version, arch, operating_system)
            req = urllib.request.Request(index_url, headers=self.headers)
            response = urllib.request.urlopen(req)
            content = response.read().decode("utf-8")
            match = re.search(r'"([^"]+\.(?:zip|tar\.gz|tar|7z))"', content)
            if match:
                filename = match.group(1)
                return True, urllib.parse.urljoin(index_url, filename)
            else:
                return (
                    False,
                    f"Failed to find any archive file (*.zip, *.tar.gz, *.tar, or *.7z) at {index_url}.",
                )

    def _get_normalized_archive_ext(self, file: str) -> SCAProcedureResult:
        if file.endswith(_TAR):
            return True, _TAR
        elif file.endswith(_TAR_GZ):
            return True, _TAR_GZ
        elif file.endswith(_ZIP):
            return True, _ZIP
        elif file.endswith(_SEVEN_ZIP):
            return True, _SEVEN_ZIP
        else:
            return False, f"Error: {file} has unexpected extension."

    def _extract_files(self, file: str, file_ending: str, destination_folder: str) -> SCAProcedureResult:
        if not os.path.isfile(file):
            return False, f"Error: {file} is not a regular file."

        start_listing = set(os.listdir(destination_folder))

        if file_ending == _TAR:
            with tarfile.open(file, "r:") as tar:
                tar.extractall(path=destination_folder)
        elif file_ending == _TAR_GZ:
            with tarfile.open(file, "r:gz") as tar:
                tar.extractall(path=destination_folder)
        elif file_ending == _ZIP:
            with zipfile.ZipFile(file, "r") as z:
                z.extractall(path=destination_folder)
        elif file_ending == _SEVEN_ZIP:
            with lzma.open(file, "rb") as z:
                with open(destination_folder, "wb") as destination:
                    shutil.copyfileobj(z, destination)

        end_listing = set(os.listdir(destination_folder))
        unzipped_directory = end_listing.difference(start_listing).pop()

        return True, os.path.join(destination_folder, unzipped_directory)

    def _path_parse(self, file_path: str) -> _Path:
        dirname = os.path.dirname(file_path)
        base = os.path.basename(file_path)
        name, ext = os.path.splitext(base)
        return _Path(dir=dirname, base=base, name=name, ext=ext)

    def _unpack_jars(self, fs_path: str, java_bin_path: str) -> SCAProcedureResult:
        if os.path.isdir(fs_path):
            for f in os.listdir(fs_path):
                current_path = os.path.join(fs_path, f)
                self._unpack_jars(current_path, java_bin_path)
            return True, None
        elif os.path.isfile(fs_path):
            file_ext = os.path.splitext(fs_path)[-1]
            if file_ext.endswith("pack"):
                p = self._path_parse(fs_path)
                name = os.path.join(p.dir, p.name)
                tool_path = os.path.join(java_bin_path, _UNPACK200)
                try:
                    subprocess.run(
                        [tool_path, _UNPACK200_ARGS, f"{name}.pack", f"{name}.jar"],
                        check=True,
                        capture_output=True,
                    )
                except (FileNotFoundError, subprocess.CalledProcessError) as e:
                    return False, str(e)
            return True, None
        else:
            return False, f"Error: {fs_path} is neither a directory not a file."

    def _decompress_archive(
        self, archive_path: str, file_ending: str, destination_folder: str
    ) -> SCAProcedureResult:
        if not os.path.isdir(destination_folder):
            os.mkdir(destination_folder)

        archive_path = os.path.normpath(archive_path)

        if os.path.isfile(archive_path):
            sucess, err_msg = self._extract_files(archive_path, file_ending, destination_folder)
            if not sucess:
                return sucess, err_msg
            else:
                unzipped_directory = err_msg
            return True, unzipped_directory
        elif os.path.isdir(archive_path):
            return True, archive_path
        else:
            return False, f"Error: {archive_path} is neither a directory not a file."

    def _get_java_filename(self, download_url) -> SCAProcedureResult:
        if self.use_chinese_jdk_mirror:
            filename = urllib.parse.urlparse(download_url).path.rpartition("/")[-1]
        else:
            req = urllib.request.Request(download_url, headers=self.headers)
            response = urllib.request.urlopen(req)
            info = response.info()
            download_url = response.geturl()
            if "Content-Disposition" not in info:
                return (
                    False,
                    f"Parsing the response from {download_url} failed. \nReason:"
                    ' "Content-Disposition" not in response.info().',
                )
            m = email.message.Message()
            m["content-type"] = info["Content-Disposition"]
            if m.get_param("filename") is None:
                return (
                    False,
                    f"Parsing the response from {download_url} failed.\nReason: can't detect the filename.",
                )
            filename = m.get_param("filename")
            if not isinstance(filename, str):
                return (
                    False,
                    f"Parsing the response from {download_url} failed.\nReason: the value of"
                    ' the attribute "filename" is not a str.',
                )
        return True, filename

    def _format_bytes(self, size: Union[int, float], precesion: int = 2):
        power = 2**10  # 1024
        n = 0
        power_labels = {0: "", 1: "K", 2: "M", 3: "G", 4: "T"}
        while size > power:
            size /= power
            n += 1
        return round(size, precesion), f"{power_labels[n]}"

    def _callbackfunc(self, block_num: int, block_size: int, total_size):
        max_hash_num = 50  # print up to 50 hashes
        precesion = 2
        downloaded_size = block_num * block_size
        if downloaded_size >= total_size:
            downloaded_size = total_size
        percent = int(100 * downloaded_size / total_size)
        downloaded_size, downloaded_size_unit = self._format_bytes(downloaded_size, precesion)
        total_size, total_size_unit = self._format_bytes(total_size, precesion)
        hash_num = int(percent / 100 * max_hash_num)
        s = (
            f"{percent:3}% [{'#' * hash_num}{' '*(max_hash_num-hash_num)}]"
            f" {downloaded_size:6} {downloaded_size_unit}/{total_size} {total_size_unit}"
        )
        same_line_print(s, width=100)

    def _download(self, download_url: str, name: str) -> SCAProcedureResult:
        if name == JAVA:
            success, err_msg = self._get_java_filename(download_url)
            if not success:
                return success, err_msg
            else:
                filename = err_msg  # e.g., jdk-18.0.2
        else:
            filename = urllib.parse.urlparse(download_url).path.rpartition("/")[-1]
            # e.g. stanford-tregex-4.2.0.zip, stanford-parser-4.2.0.zip
        filename = os.path.join(tempfile.gettempdir(), filename)  # type: ignore
        try:
            opener = urllib.request.build_opener()
            opener.addheaders = list(self.headers.items())
            urllib.request.install_opener(opener)
            urllib.request.urlretrieve(download_url, filename, self._callbackfunc)
            same_line_print("", width=100)
        except URLError as e:
            if hasattr(e, "reason"):
                return (False, f"Requesting to {download_url} failed.\nReason: {e.reason}")
            elif hasattr(e, "code"):
                return (False, f"Requesting to {download_url} failed.\nReason: {e.code}")
            else:
                return False, f"Requesting to {download_url} failed."
        return True, filename

    def ask_install(self, name: str, assume_yes: bool = False) -> SCAProcedureResult:
        reason_dict = {
            JAVA: f"values of PATH does not contain a {JAVA} bin folder",
            STANFORD_PARSER: f"the environment variable STANFORD_PARSER_HOME is not found",
            STANFORD_TREGEX: f"the environment variable STANFORD_TREGEX_HOME is not found",
        }
        if assume_yes:
            is_install = "y"
        else:
            is_install = get_yes_or_no(
                f"It seems that {name} has not been installed, because {reason_dict[name]}. Do you want to"
                " let NeoSCA install it for you?"
            )
        if is_install in ("n", "N"):
            manual_install_prompt_dict = {
                JAVA: (
                    f"You will have to install {JAVA} manually.\n\n1. To install it, visit"
                    " https://www.java.com/en/download.\n2. After installing, make sure you can access it in"
                    " the cmd window by typing in `java -version`."
                ),
                STANFORD_PARSER: (
                    f"You will have to install {STANFORD_PARSER} manually.\n\n1. To install it,"
                    f" download and unzip the archive file at {self._URL_STANFORD_PARSER}.\n2. Set an"
                    " environment variable STANFORD_PARSER_HOME to the path of the unzipped"
                    " directory."
                ),
                STANFORD_TREGEX: (
                    f"You will have to install {STANFORD_TREGEX} manually.\n\n1. To install it,"
                    f" download and unzip the archive file at {self._URL_STANFORD_TREGEX}.\n2. Set an"
                    " environment variable STANFORD_PARSER_HOME to the path of the unzipped"
                    " directory."
                ),
            }
            return (False, manual_install_prompt_dict[name])
        else:
            return True, None

    def install_java(
        self,
        version: str,
        operating_system: str,
        arch: str,
        impl: str,
        target_dir: str,
        assume_yes: bool = False,
    ) -> SCAProcedureResult:
        match = glob.glob(f"{target_dir}{os.sep}jdk{version}*")
        if match:
            return True, match[0]
        sucess, err_msg = self.ask_install(JAVA, assume_yes)
        if not sucess:
            return sucess, err_msg
        sucess, err_msg = self.get_java_download_url(version, operating_system, arch, impl)
        if not sucess:
            return sucess, err_msg
        else:
            url = err_msg
        color_print("OKGREEN", target_dir, prefix=f'Downloading {url} to "', postfix='"...')
        sucess, err_msg = self._download(url, name=JAVA)  # type:ignore
        if not sucess:
            return sucess, err_msg
        jdk_archive = err_msg
        sucess, err_msg = self._get_normalized_archive_ext(jdk_archive)  # type:ignore
        if not sucess:
            return sucess, err_msg
        jdk_ext = err_msg
        print(f"Decompressing {JAVA} archive...")
        sucess, err_msg = self._decompress_archive(jdk_archive, jdk_ext, target_dir)  # type:ignore
        if not sucess:
            return sucess, err_msg
        jdk_dir = err_msg
        jdk_bin = os.path.join(jdk_dir, "bin")  # type:ignore
        sucess, err_msg = self._unpack_jars(jdk_dir, jdk_bin)  # type:ignore
        if not sucess:
            return sucess, err_msg
        if jdk_archive:
            os.remove(jdk_archive)
        return True, jdk_dir

    def install_stanford(
        self, name: str, url: str, target_dir: str, assume_yes: bool = False
    ) -> SCAProcedureResult:
        match = glob.glob(f"{target_dir}{os.sep}{name.lower().replace(' ', '-')}*")
        if match:
            return True, match[0]
        sucess, err_msg = self.ask_install(name, assume_yes)
        if not sucess:
            return sucess, err_msg
        color_print("OKGREEN", target_dir, prefix=f'Downloading {url} to "', postfix='"...')
        sucess, err_msg = self._download(url, name=name)
        if not sucess:
            return sucess, err_msg
        archive_file = err_msg
        sucess, err_msg = self._get_normalized_archive_ext(archive_file)  # type:ignore
        if not sucess:
            return sucess, err_msg
        archive_ext = err_msg
        print(f"Decompressing {name} archive...")
        sucess, err_msg = self._decompress_archive(archive_file, archive_ext, target_dir)  # type:ignore
        if not sucess:
            return sucess, err_msg
        unzipped_directory = err_msg
        if archive_file:
            os.remove(archive_file)
        return True, unzipped_directory

    def install(
        self,
        name: str,
        assume_yes: bool = False,
        version: str = _JAVA_VERSION,
        operating_system: str = OS,
        arch: str = ARCH,
        impl: str = Implementation.HOTSPOT,
        target_dir: str = _TARGET_DIR,  # type: ignore
    ) -> SCAProcedureResult:
        if name == JAVA:
            return self.install_java(version, operating_system, arch, impl, target_dir, assume_yes)
        elif name == STANFORD_PARSER:
            return self.install_stanford(name, self._URL_STANFORD_PARSER, target_dir, assume_yes)
        elif name == STANFORD_TREGEX:
            return self.install_stanford(name, self._URL_STANFORD_TREGEX, target_dir, assume_yes)
        else:
            return False, f"Unexpected name: {name}."



import sys
from typing import Union, Sequence


class Structure:
    def __init__(self, name: str, desc: str, pat="", matches="") -> None:
        """
        :param name: name of the structure
        :param desc: description of the structure
        :param pat: Tregex pattern
        :param matches: matched subtrees by Tregex
        """
        self.name = name
        self.desc = desc
        self.pat = pat
        self.matches = matches
        self.freq: Union[float, int] = 0


class Structures:
    W = Structure("W", "words")
    S = Structure("S", "sentences", "ROOT")
    VP1 = Structure("VP1", "regular verb phrases", "VP > S|SINV|SQ")
    VP2 = Structure(
        "VP2",
        "verb phrases in inverted yes/no questions or in wh-questions",
        "MD|VBZ|VBP|VBD > (SQ !< VP)",
    )
    C1 = Structure(
        "C1",
        "regular clauses",
        "S|SINV|SQ "
        "[> ROOT <, (VP <# VB) | <# MD|VBZ|VBP|VBD | < "
        "(VP [<# MD|VBP|VBZ|VBD | < CC < (VP <# MD|VBP|VBZ|VBD)])]",
    )
    C2 = Structure(
        "C2",
        "fragment clauses",
        "FRAG > ROOT !<< "
        "(S|SINV|SQ [> ROOT <, (VP <# VB) | <# MD|VBZ|VBP|VBD | < "
        "(VP [<# MD|VBP|VBZ|VBD | < CC < "
        "(VP <# MD|VBP|VBZ|VBD)])])",
    )
    T1 = Structure(
        "T",
        "regular T-units",
        "S|SBARQ|SINV|SQ > ROOT | [$-- S|SBARQ|SINV|SQ !>> SBAR|VP]",
    )
    T2 = Structure(
        "T2",
        "fragment T-units",
        "FRAG > ROOT !<< (S|SBARQ|SINV|SQ > ROOT | [$-- S|SBARQ|SINV|SQ !>> SBAR|VP])",
    )
    CN1 = Structure(
        "CN1",
        "complex nominals, type 1",
        "NP !> NP [<< JJ|POS|PP|S|VBG | << (NP $++ NP !$+ CC)]",
    )
    CN2 = Structure(
        "CN2",
        "complex nominals, type 2",
        "SBAR [<# WHNP | <# (IN < That|that|For|for) | <, S] & [$+ VP | > VP]",
    )
    CN3 = Structure("CN3", "complex nominals, type 3", "S < (VP <# VBG|TO) $+ VP")
    DC = Structure(
        "DC",
        "dependent clauses",
        "SBAR < "
        "(S|SINV|SQ "
        "[> ROOT <, (VP <# VB) | <# MD|VBZ|VBP|VBD | < "
        "(VP [<# MD|VBP|VBZ|VBD | < CC < (VP <# MD|VBP|VBZ|VBD)])])",
    )
    CT = Structure(
        "CT",
        "complex T-units",
        "S|SBARQ|SINV|SQ [> ROOT | [$-- S|SBARQ|SINV|SQ !>> SBAR|VP]] << "
        "(SBAR < (S|SINV|SQ "
        "[> ROOT <, (VP <# VB) | <# MD|VBZ|VBP|VBD | < "
        "(VP [<# MD|VBP|VBZ|VBD | < CC < "
        "(VP <# MD|VBP|VBZ|VBD)])]))",
    )
    CP = Structure("CP", "coordinate phrases", "ADJP|ADVP|NP|VP < CC")

    VP = Structure("VP", "verb phrases")
    C = Structure("C", "clauses")
    T = Structure("T", "T-units")
    CN = Structure("CN", "complex nominals")

    MLS = Structure("MLS", "mean length of sentence")
    MLT = Structure("MLT", "mean length of T-unit")
    MLC = Structure("MLC", "mean length of clause")
    CpS = Structure("C/S", "clauses per sentence")
    VPpT = Structure("VP/T", "verb phrases per T-unit")
    CpT = Structure("C/T", "clauses per T-unit")
    DCpC = Structure("DC/C", "dependent clauses per clause")
    DCpT = Structure("DC/T", "dependent clauses per T-unit")
    TpS = Structure("T/S", "T-units per sentence")
    CTpT = Structure("CT/T", "complex T-unit ratio")
    CPpT = Structure("CP/T", "coordinate phrases per T-unit")
    CPpC = Structure("CP/C", "coordinate phrases per clause")
    CNpT = Structure("CN/T", "complex nominals per T-unit")
    CNpC = Structure("CN/C", "complex nominals per clause")

    # a list of tregex patterns for various structures
    to_query: Sequence[Structure] = (
        S,
        VP1,
        VP2,
        C1,
        C2,
        T1,
        T2,
        CN1,
        CN2,
        CN3,
        DC,
        CT,
        CP,
    )
    to_report: Sequence[Structure] = (
        S,
        VP,
        C,
        T,
        DC,
        CT,
        CP,
        CN,
        MLS,
        MLT,
        MLC,
        CpS,
        VPpT,
        CpT,
        DCpC,
        DCpT,
        TpS,
        CTpT,
        CPpT,
        CPpC,
        CNpT,
        CNpC,
    )

    fields = "Filename,W," + ",".join((structure.name for structure in to_report))

    def __init__(self, ifile) -> None:
        self.ifile = ifile

    def update_freqs(self) -> None:
        """
        Update frequencies of complex nominals, clauses, verb phrases, and T-units
        """
        self.CN.freq = self.CN1.freq + self.CN2.freq + self.CN3.freq
        self.C.freq = self.C1.freq + self.C2.freq
        self.VP.freq = self.VP1.freq + self.VP2.freq
        self.T.freq = self.T1.freq + self.T2.freq

    def _div(self, x, y) -> Union[float, int]:
        return round(x / y, 4) if y else 0

    def compute_14_indicies(self) -> None:
        """
        Compute the 14 syntactic complexity indices
        """
        self.MLS.freq = self._div(self.W.freq, self.S.freq)
        self.MLT.freq = self._div(self.W.freq, self.T.freq)
        self.MLC.freq = self._div(self.W.freq, self.C1.freq)
        self.CpS.freq = self._div(self.C1.freq, self.S.freq)
        self.VPpT.freq = self._div(self.VP1.freq, self.T.freq)
        self.CpT.freq = self._div(self.C1.freq, self.T.freq)
        self.DCpC.freq = self._div(self.DC.freq, self.C1.freq)
        self.DCpT.freq = self._div(self.DC.freq, self.T.freq)
        self.TpS.freq = self._div(self.T.freq, self.S.freq)
        self.CTpT.freq = self._div(self.CT.freq, self.T.freq)
        self.CPpT.freq = self._div(self.CP.freq, self.T.freq)
        self.CPpC.freq = self._div(self.CP.freq, self.C1.freq)
        self.CNpT.freq = self._div(self.CN.freq, self.T.freq)
        self.CNpC.freq = self._div(self.CN.freq, self.C1.freq)

    def get_freqs(self) -> dict:
        freq_dict = {"Filename": self.ifile, "W": self.W.freq}
        for structure in self.to_report:
            freq_dict[structure.name] = structure.freq
        return freq_dict



#!/usr/bin/env python3
# -*- coding=utf-8 -*-

import re
import os
import sys
import subprocess
from typing import Tuple
from .structures import Structure


class Querier:
    TREGEX_METHOD = "edu.stanford.nlp.trees.tregex.TregexPattern"
    MAX_MEMORY = "100m"

    def __init__(self, dir_stanford_tregex: str, max_memory=MAX_MEMORY) -> None:
        self.classpath = dir_stanford_tregex + os.sep + "stanford-tregex.jar"
        self.max_memory = max_memory

    def query(self, structure: Structure, trees: str) -> Tuple[int, str]:
        """Call Tregex to query {pattern} against {ofile_parsed}"""
        print(f'\t[Tregex] Querying "{structure.desc}"...')
        cmd = [
            "java",
            f"-mx{self.max_memory}",
            "-cp",
            self.classpath,
            self.TREGEX_METHOD,
            structure.pat,
            "-o",
            "-filter",
        ]
        try:
            p = subprocess.run(
                cmd,
                input=trees.encode("utf-8"),
                check=True,
                capture_output=True,
            )
        except subprocess.CalledProcessError as err_msg:
            print(err_msg)
            sys.exit(1)
        match_reslt = re.search(r"There were (\d+) matches in total\.", p.stderr.decode("utf-8"))
        if match_reslt:
            freq = match_reslt.group(1)
        else:
            sys.exit(
                "Error: failed to obtain frequency. It is likely that:\n"
                "(1) Tregex's interface has"
                " changed. As a workaround, download an older version, v4.2.0,"
                " for example. If Tregex's interface does have changed, the"
                " latest version of Tregex will be supported in next few"
                " releases.\n"
                "(2) You manually modified Tregex patterns in"
                " structures.py. Make sure that: on Windows, those patterns are"
                " enclosed by double quotes; on Linux and macOS they are"
                " surrounded by single quotes."
            )
        matched_subtrees = p.stdout.decode("utf-8")
        matched_subtrees = self._add_terms(matched_subtrees)
        return int(freq), matched_subtrees

    def _add_terms(self, subtrees: str) -> str:
        """
        Add terminals above each subtree

        :param subtrees: matched subtrees by Tregex
         e.g. (NP (NNP Mr.) (NNP Reed))
        :return: e.g.
         Mr. Reed
         (NP (NNP Mr.) (NNP Reed))
        """
        result = ""
        pattern = r"([^\s)]+)\)"
        for subtree in re.split(r"(?:\r?\n){2,}", subtrees):
            terminals = " ".join(re.findall(pattern, subtree))
            result += f"{terminals}\n{subtree}\n\n"
        return result.strip()



#!/usr/bin/env python3
# -*- coding=utf-8 -*-

import os
import re
import sys
from typing import List, Optional, Tuple

# For all the procedures in SCAUI, return a tuple as the result
# The first element bool indicates whether the procedure succeeds
# The second element is the error message if it fails.
SCAProcedureResult = Tuple[bool, Optional[str]]


class _bcolors:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


bcolors = _bcolors()
color_support = True
if sys.platform == "win32":
    try:
        # https://stackoverflow.com/questions/36760127/...
        # how-to-use-the-new-support-for-ansi-escape-sequences-in-the-windows-10-console
        from ctypes import windll

        kernel32 = windll.kernel32
        kernel32.SetConsoleMode(kernel32.GetStdHandle(-11), 7)
    except Exception:  # pragma: no cover
        color_support = False


def color_print(color: str, s: str, prefix: str = "", postfix: str = "", **kwargs) -> None:
    kwargs.update({"sep": ""})
    if color_support:
        print(prefix, bcolors.__getattribute__(color) + s + bcolors.ENDC, postfix, **kwargs)
    else:  # pragma: no cover
        print(prefix, s, postfix, **kwargs)


def same_line_print(s: str, width=80, **kwargs) -> None:
    print(f"\r{'':<{width}}", end="")  # clear the line
    print(f"\r{s}", end="", **kwargs)


def try_write(filename: str, content: Optional[str]) -> SCAProcedureResult:
    try:
        with open(filename, "w", encoding="utf-8") as f:
            if content is not None:
                f.write(content)
            return True, None
    except PermissionError:
        return (
            False,
            f"PermissionError: can not write to {filename}, because it is already"
            f" in use by another process.\n\n1. Ensure that {filename} is closed,"
            " or \n2. Specify another output filename through the `-o` option,"
            f" e.g. nsca input.txt -o {filename.replace('.csv', '-2.csv')}",
        )


def _setenv_windows(env_var: str, paths: List[str], refresh: bool = False) -> None:
    import winreg  # Allows access to the windows registry
    import ctypes  # Allows interface with low-level C API's

    with winreg.ConnectRegistry(None, winreg.HKEY_CURRENT_USER) as root:  # type:ignore
        # Get the current user registry
        with winreg.OpenKey(root, "Environment", 0, winreg.KEY_ALL_ACCESS) as key:  # type:ignore
            # Go to the environment key
            if refresh or os.environ.get(env_var) is None:
                new_value = ";".join(paths)
            else:
                existing_value = os.environ.get(env_var)
                new_value = existing_value + ";" + ";".join(paths)  # type:ignore
                # Takes the current path value and appends the new program path
            winreg.SetValueEx(key, env_var, 0, winreg.REG_EXPAND_SZ, new_value)  # type:ignore
            # Updated the path with the updated path

        # Tell other processes to update their environment
        HWND_BROADCAST = 0xFFFF
        WM_SETTINGCHANGE = 0x1A
        SMTO_ABORTIFHUNG = 0x0002
        result = ctypes.c_long()
        SendMessageTimeoutW = ctypes.windll.user32.SendMessageTimeoutW  # type:ignore
        SendMessageTimeoutW(
            HWND_BROADCAST,
            WM_SETTINGCHANGE,
            0,
            "Environment",
            SMTO_ABORTIFHUNG,
            5000,
            ctypes.byref(result),
        )


def _setenv_unix(env_var: str, paths: List[str], refresh: bool = False) -> None:
    shell = os.environ.get("SHELL")
    if shell is None:
        print("Failed to permanently append {path} to {env_var}.\nReason: can't detect current shell.")
    else:
        startup_file_dict = {
            "bash": "~/.bash_profile" if sys.platform == "darwin" else "~/.bashrc",
            "zsh": "~/.zshrc",
            "ksh": "~/.kshrc",
            "tcsh": "~/.tcshrc",
            "csh": "~/.cshrc",
            "yash": "~/.yashrc",
            "fish": "~/.config/fish/config.fish",
            "ion": "~/.config/ion/initrc",
        }
        startup_file = startup_file_dict.get(os.path.basename(shell), None)
        if startup_file is None:
            print(
                f"Failed to permanently set environment variables.\nReason: can't detect rc file for {shell}."
            )
        else:
            new_paths = '"' + '":"'.join(paths) + '"'
            startup_file = os.path.expanduser(startup_file)
            with open(startup_file, "r", encoding="utf-8") as f:
                configs = [line.strip() for line in f.readlines()]
            new_config = (
                f"export {env_var}={new_paths}" if refresh else f"export {env_var}=${env_var}:{new_paths}"
            )
            duplicated_config_index = []
            for i, config in enumerate(configs):
                if refresh:
                    if config.startswith(f"export {env_var}"):
                        duplicated_config_index.append(i)
                else:
                    if config == new_config:
                        duplicated_config_index.append(i)
            for i in duplicated_config_index:
                del configs[i]
            if duplicated_config_index:
                configs.insert(duplicated_config_index[0], new_config)
            else:
                configs.append(new_config)
            with open(startup_file, "w", encoding="utf-8") as f:
                f.write("\n".join(configs))


def setenv(env_var: str, paths: List[str], refresh: bool = False) -> None:
    assert sys.platform in ("win32", "darwin", "linux")
    if sys.platform == "win32":
        _setenv_windows(env_var, paths, refresh)
    else:
        _setenv_unix(env_var, paths, refresh)
    color_print(
        "OKGREEN",
        env_var,
        prefix="Added the following path(s) to ",
        postfix=":\n" + "\n".join(paths),
        end="\n\n",
    )


def get_yes_or_no(prompt: str = "") -> str:
    prompt_options = "Enter [y]es or [n]o: "
    sep = "\n" if prompt else ""
    answer = input(prompt + sep + prompt_options)
    while answer not in ("y", "n", "Y", "N"):
        answer = input(f"Unexpected input: {answer}.\nEnter [y]es or [n]o: ")
    return answer



import os
import re
import sys
from typing import Tuple, Generator

from .parser import Parser
from .querier import Querier
from .structures import Structures


class NeoSCA:
    def __init__(
        self,
        dir_stanford_parser: str,
        dir_stanford_tregex: str,
        reserve_parsed: bool,
        verbose: bool = True,
    ) -> None:
        self.parser = Parser(dir_stanford_parser, verbose=verbose)
        self.querier = Querier(dir_stanford_tregex)

        self.reserve_parsed = reserve_parsed
        self.skip_parsing = False

    def _parse_ifile(self, ifile: str, ofile_parsed: str) -> str:
        """Parse a single text file"""
        if os.path.exists(ofile_parsed) and os.path.getsize(ofile_parsed) > 0:
            mt_input = os.path.getmtime(ifile)  # get the last modification time
            mt_parsed = os.path.getmtime(ofile_parsed)
            if mt_input < mt_parsed:
                self.skip_parsing = True
                print(
                    f"\t[Parser] Parsing skipped: {ofile_parsed} already"
                    f" exists, and is non-empty and newer than {ifile}."
                )
        if not self.skip_parsing:
            with open(ifile, "r", encoding="utf-8") as f:
                text = f.read()
            trees = self.parser.parse(text, ofile_parsed)
        else:
            with open(ofile_parsed, "r", encoding="utf-8") as f:
                trees = f.read()

        if self.reserve_parsed:
            with open(ofile_parsed, "w", encoding="utf-8") as f:
                f.write(trees)
        return trees

    def _query_against_trees(self, trees: str, structures: Structures) -> Structures:
        for structure in structures.to_query:
            structure.freq, structure.matches = self.querier.query(structure, trees)

        structures.W.freq = len(re.findall(r"\([A-Z]+\$? [^()]+\)", trees))
        structures.update_freqs()
        structures.compute_14_indicies()
        return structures

    def parse_text(self, text: str, ofile_parsed="cmdline_text.parsed") -> str:
        trees = self.parser.parse(text, ofile_parsed)
        if self.reserve_parsed:
            with open(ofile_parsed, "w", encoding="utf-8") as f:
                f.write(trees)
        return trees

    def parse_text_and_query(self, text: str) -> Structures:
        trees = self.parse_text(text)
        structures = Structures("cmdline_text")
        return self._query_against_trees(trees, structures)

    def parse_ifiles_and_query(self, ifiles) -> Generator[Structures, None, None]:
        total = len(ifiles)
        for i, ifile in enumerate(ifiles):
            print(f'[NeoSCA] Processing "{ifile}" ({i+1}/{total})...')
            ofile_parsed = os.path.splitext(ifile)[0] + ".parsed"
            try:
                structures = Structures(ifile)
                trees = self._parse_ifile(ifile, ofile_parsed)
                yield self._query_against_trees(trees, structures)
            except KeyboardInterrupt:
                if os.path.exists(ofile_parsed) and not self.skip_parsing:
                    os.remove(ofile_parsed)
                sys.exit(1)

    def parse_ifiles(self, ifiles) -> None:
        total = len(ifiles)
        for i, ifile in enumerate(ifiles):
            print(f'[NeoSCA] Processing "{ifile}" ({i+1}/{total})...')
            ofile_parsed = os.path.splitext(ifile)[0] + ".parsed"
            try:
                self._parse_ifile(ifile, ofile_parsed)
            except KeyboardInterrupt:
                if os.path.exists(ofile_parsed) and not self.skip_parsing:
                    os.remove(ofile_parsed)
                sys.exit(1)



#!/usr/bin/env python3
# -*- coding=utf-8 -*-
import os
import subprocess
import sys


class Parser:
    PARSER_METHOD = "edu.stanford.nlp.parser.lexparser.LexicalizedParser"
    PARSER_MODEL = "edu/stanford/nlp/models/lexparser/englishPCFG.ser.gz"
    NTHREADS = "2"
    MAX_MEMORY = "3g"

    def __init__(
        self,
        dir_stanford_parser: str,
        nthreads=NTHREADS,
        max_memory=MAX_MEMORY,
        verbose=False,
    ) -> None:
        self.classpath = dir_stanford_parser + os.sep + "*"
        self.max_memory = max_memory
        self.nthreads = nthreads
        self.verbose = verbose

    def parse(self, text: str, ofile_parsed: str) -> str:
        """Call Stanford Parser"""
        cmd = [
            "java",
            f"-mx{self.max_memory}",
            "-cp",
            self.classpath,
            self.PARSER_METHOD,
            "-outputFormat",
            "penn",
            "-nthreads",
            self.nthreads,
            self.PARSER_MODEL,
            "-",
        ]
        print(f"\t[Parser] Parsing with command ", " ".join(cmd), "...", sep="")
        try:
            p = subprocess.run(
                cmd,
                input=text.encode("utf-8"),
                check=True,
                stdout=subprocess.PIPE,
                stderr=None if self.verbose else subprocess.DEVNULL,
            )
        except (FileNotFoundError, subprocess.CalledProcessError) as err_msg:
            print(err_msg)
            if os.path.exists(ofile_parsed):
                os.remove(ofile_parsed)
            sys.exit(1)
        return p.stdout.decode("utf-8")



import setuptools

with open("./README.md", "r", encoding="utf-8") as f:
    long_description = f.read()
with open("./neosca/__init__.py", "r", encoding="utf-8") as f:
    for line in f.readlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            version = line.split(delim)[1]
            break
    else:
        print("Can't find version! Stop Here!")
        exit(1)
setuptools.setup(
    name="neosca",
    version=version,
    author="TAN Long",
    author_email="tanloong@foxmail.com",
    url="https://github.com/tanloong/neosca",
    packages=["neosca"],
    description=(
        "Another syntactic complexity analyzer of written English language samples"
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[],
    classifiers=[
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.7",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v2 or later"
        " (GPLv2+)",
        "Operating System :: MacOS",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows",
    ],
    python_requires=">=3.7",
    entry_points={"console_scripts": ["nsca = neosca:main"]},
)



import argparse
import glob
from itertools import tee
from os import path
import os
import subprocess
import sys
from typing import List

from . import __version__
from .neosca import NeoSCA
from .writer import write_match_output
from .writer import write_freq_output
from .util import SCAProcedureResult
from .util import color_print
from .util import try_write


class SCAUI:
    def __init__(self) -> None:
        self.args_parser: argparse.ArgumentParser = self.create_args_parser()
        self.options: argparse.Namespace = argparse.Namespace()
        self.cwd = os.getcwd()
        self.STANFORD_PARSER_HOME = "STANFORD_PARSER_HOME"
        self.STANFORD_TREGEX_HOME = "STANFORD_TREGEX_HOME"

    def create_args_parser(self) -> argparse.ArgumentParser:
        args_parser = argparse.ArgumentParser(prog="nsca")
        args_parser.add_argument(
            "--version",
            action="store_true",
            default=False,
            help="show the version of NeoSCA",
        )
        args_parser.add_argument(
            "--list",
            dest="list_fields",
            action="store_true",
            default=False,
            help="list the available output fields",
        )
        args_parser.add_argument(
            "--text",
            "-t",
            default=None,
            help="pass text through the command line",
        )
        args_parser.add_argument(
            "--output-file",
            "-o",
            metavar="OUTFILE",
            dest="ofile_freq",
            default=None,
            help="specify an output file",
        )
        args_parser.add_argument(
            "--output-format",
            dest="oformat_freq",
            choices=["csv", "json"],
            default="csv",
            help="output format, the default is csv",
        )
        args_parser.add_argument(
            "--stdout",
            dest="stdout",
            action="store_true",
            default=False,
            help="write the frequency output to the stdout instead of saving it to a file",
        )
        args_parser.add_argument(
            "--reserve-parsed",
            "-p",
            dest="reserve_parsed",
            action="store_true",
            default=False,
            help="reserve the parsed trees produced by the Stanford Parser",
        )
        args_parser.add_argument(
            "--reserve-matched",
            "-m",
            dest="reserve_matched",
            default=False,
            action="store_true",
            help="reserve the matched subtrees produced by the Stanford Tregex",
        )
        args_parser.add_argument(
            "--no-query",
            dest="no_query",
            action="store_true",
            default=False,
            help="parse the input files, save the parsed trees, and exit",
        )
        args_parser.add_argument(
            "--check-depends",
            dest="check_depends",
            action="store_true",
            default=False,
            help="check NeoSCA's dependencies, including Java, Stanford Parser, and Stanford Tregex",
        )
        args_parser.add_argument(
            "--yes",
            dest="assume_yes",
            action="store_true",
            default=False,
            help="assume the answer to all prompts is yes, used when installing dependencies",
        )
        return args_parser

    def parse_args(self, argv: List[str]) -> SCAProcedureResult:
        options, ifile_list = self.args_parser.parse_known_args(argv[1:])
        if options.no_query:
            options.reserve_parsed = True

        if options.text is None:
            verified_ifile_list = []
            for f in ifile_list:
                if path.isfile(f):
                    verified_ifile_list.append(f)
                elif glob.glob(f):
                    verified_ifile_list.extend(glob.glob(f))
                else:
                    return (False, f"No such file as \n\n{f}")
        else:
            print(f"Command-line text: {options.text}")
            verified_ifile_list = None
        self.verified_ifile_list = verified_ifile_list

        self.odir_match = "result_matches"
        if options.stdout:
            options.ofile_freq = sys.stdout
        elif options.ofile_freq is not None:
            self.odir_match = os.path.splitext(options.ofile_freq)[0] + "_matches"
            ofile_freq_ext = os.path.splitext(options.ofile_freq)[-1].lstrip(".")
            if ofile_freq_ext not in ("csv", "json"):
                return (
                    False,
                    f"The file extension {ofile_freq_ext} is not supported. Use one of"
                    " the following:\n1. csv\n2. json",
                )
            if ofile_freq_ext != options.oformat_freq:
                options.oformat_freq = ofile_freq_ext
        else:
            options.ofile_freq = "result." + options.oformat_freq

        self.init_kwargs = {
            "dir_stanford_parser": "",
            "dir_stanford_tregex": "",
            "reserve_parsed": options.reserve_parsed,
        }
        self.options = options
        return True, None

    def check_java(self) -> SCAProcedureResult:
        try:
            subprocess.run(["java", "-version"], check=True, capture_output=True)
        except (FileNotFoundError, subprocess.CalledProcessError):
            from .depends_installer import depends_installer
            from .depends_installer import JAVA
            from .util import setenv

            installer = depends_installer()
            sucess, err_msg = installer.install(JAVA, assume_yes=self.options.assume_yes)
            if not sucess:
                return sucess, err_msg
            else:
                java_home = err_msg
                java_bin = os.path.join(java_home, "bin")  # type:ignore
                setenv("JAVA_HOME", [java_home], True)  # type:ignore
                setenv("PATH", [java_bin], False)  # type:ignore
                current_PATH = os.environ.get("PATH", default="")
                os.environ["PATH"] = current_PATH + os.pathsep + java_bin  # type:ignore
        else:
            print("Java has already been installed.")
        return True, None

    def check_stanford_parser(self) -> SCAProcedureResult:
        try:
            self.options.dir_stanford_parser = os.environ[self.STANFORD_PARSER_HOME]
        except KeyError:
            from .depends_installer import depends_installer
            from .depends_installer import STANFORD_PARSER
            from .util import setenv

            installer = depends_installer()
            sucess, err_msg = installer.install(STANFORD_PARSER, assume_yes=self.options.assume_yes)
            if not sucess:
                return sucess, err_msg
            else:
                stanford_parser_home = err_msg
                setenv(self.STANFORD_PARSER_HOME, [stanford_parser_home], True)  # type:ignore
                self.options.dir_stanford_parser = stanford_parser_home  # type:ignore
        else:
            print("Stanford Parser has already been installed.")
        self.init_kwargs.update({"dir_stanford_parser": self.options.dir_stanford_parser})
        return True, None

    def check_stanford_tregex(self) -> SCAProcedureResult:
        try:
            self.options.dir_stanford_tregex = os.environ[self.STANFORD_TREGEX_HOME]
        except KeyError:
            from .depends_installer import depends_installer
            from .depends_installer import STANFORD_TREGEX
            from .util import setenv

            installer = depends_installer()
            sucess, err_msg = installer.install(STANFORD_TREGEX, assume_yes=self.options.assume_yes)
            if not sucess:
                return sucess, err_msg
            else:
                stanford_tregex_home = err_msg
                setenv(self.STANFORD_TREGEX_HOME, [stanford_tregex_home], True)  # type:ignore
                self.options.dir_stanford_tregex = stanford_tregex_home  # type:ignore
        else:
            print("Stanford Tregex has already been installed.")
        self.init_kwargs.update({"dir_stanford_tregex": self.options.dir_stanford_tregex})
        return True, None

    def check_depends(self) -> SCAProcedureResult:
        success_java, err_msg_java = self.check_java()
        success_parser, err_msg_parser = self.check_stanford_parser()
        success_tregex, err_msg_tregex = self.check_stanford_tregex()

        sucesses = (success_java, success_parser, success_tregex)
        err_msges = (err_msg_java, err_msg_parser, err_msg_tregex)
        if all(sucesses):
            return True, None
        else:
            err_msg = "\n\n".join(
                map(lambda p: p[1] if not p[0] else "", zip(sucesses, err_msges))  # type:ignore
            )
            return False, err_msg.strip()

    def check_python(self) -> SCAProcedureResult:
        v_info = sys.version_info
        if v_info.minor >= 7 and v_info.major == 3:
            return True, None
        else:
            return (
                False,
                f"Error: Python {v_info.major}.{v_info.minor} is too old."
                " NeoSCA only supports Python 3.7 or higher.",
            )

    def exit_routine(self) -> None:
        print("\n", "=" * 60, sep="")
        i = 1
        if not self.options.no_query and not self.options.stdout:
            color_print(
                "OKGREEN",
                f"{path.abspath(self.options.ofile_freq)}",
                prefix=f"{i}. Frequency output was saved to ",
                postfix=".",
            )
            i += 1
        if self.verified_ifile_list and self.options.reserve_parsed:
            print(
                f"{i}. Parsed trees were saved corresponding to input files,"
                ' with the same name but a ".parsed" extension.'
            )
            i += 1
        if self.options.text is not None and self.options.reserve_parsed:
            color_print(
                "OKGREEN",
                f"{self.cwd}{os.sep}cmdline_text.parsed",
                prefix=f"{i}. Parsed trees were saved to ",
                postfix=".",
            )
            i += 1
        if self.options.reserve_matched:
            color_print(
                "OKGREEN",
                f"{path.abspath(self.odir_match)}",
                prefix=f"{i}. Matched subtrees were saved to ",
                postfix=".",
            )
            i += 1
        print("Done.")

    def run_tmpl(func):  # type: ignore
        def wrapper(self, *args, **kwargs):
            sucess, err_msg = self.check_python()
            if not sucess:
                return sucess, err_msg
            sucess, err_msg = self.check_depends()
            if not sucess:
                return sucess, err_msg
            if not self.options.stdout:
                sucess, err_msg = try_write(self.options.ofile_freq, None)
                if not sucess:
                    return sucess, err_msg
            func(self, *args, **kwargs)  # type: ignore
            self.exit_routine()
            return True, None

        return wrapper

    @run_tmpl  # type: ignore
    def run_parse_text(self) -> None:
        analyzer = NeoSCA(**self.init_kwargs)
        analyzer.parse_text(self.options.text)

    @run_tmpl  # type: ignore
    def run_parse_text_and_query(self) -> None:
        analyzer = NeoSCA(**self.init_kwargs)
        structures = analyzer.parse_text_and_query(self.options.text)

        write_freq_output([structures], self.options.ofile_freq, self.options.oformat_freq)

        if self.options.reserve_matched:
            write_match_output(structures, self.odir_match)

    @run_tmpl  # type: ignore
    def run_parse_ifiles(self) -> None:
        analyzer = NeoSCA(**self.init_kwargs)
        analyzer.parse_ifiles(self.verified_ifile_list)

    @run_tmpl  # type: ignore
    def run_parse_ifiles_and_query(self) -> None:
        analyzer = NeoSCA(**self.init_kwargs)
        gen = analyzer.parse_ifiles_and_query(self.verified_ifile_list)
        structures_generator1, structures_generator2 = tee(gen, 2)
        # a generator of instances of Structures, each for one corresponding input file

        write_freq_output(
            structures_generator1,
            self.options.ofile_freq,
            self.options.oformat_freq,
        )

        if self.options.reserve_matched:
            for structures in structures_generator2:
                write_match_output(structures, self.odir_match)

    def run(self) -> SCAProcedureResult:
        if self.options.version:
            return self.show_version()
        elif self.options.list_fields:
            return self.list_fields()
        elif self.options.check_depends:
            return self.check_depends()
        elif self.options.text is not None and self.options.no_query:
            return self.run_parse_text()  # type: ignore
        elif self.options.text is not None and not self.options.no_query:
            return self.run_parse_text_and_query()  # type: ignore
        elif self.verified_ifile_list and self.options.no_query:
            return self.run_parse_ifiles()  # type: ignore
        elif self.verified_ifile_list and not self.options.no_query:
            return self.run_parse_ifiles_and_query()  # type: ignore
        else:
            self.args_parser.print_help()
            return True, None

    def list_fields(self) -> SCAProcedureResult:
        from .structures import Structures

        field_info = "W: words"
        for structure in Structures.to_report:
            field_info += f"\n{structure.name}: {structure.desc}"
        print(field_info)
        return True, None

    def show_version(self) -> SCAProcedureResult:
        print(__version__)
        return True, None


def main() -> None:
    ui = SCAUI()
    success, err_msg = ui.parse_args(sys.argv)
    if not success:
        print(err_msg)
        sys.exit(1)
    success, err_msg = ui.run()
    if not success:
        print(err_msg)
        sys.exit(1)
